#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2020 Huawei Technologies Co., Ltd
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import acl
import struct
import numpy as np
from ..resource.mem import *
from ..resource.context import bind_context
from ..common import const


class AscendArray():
    """Define a AscendArray data class like numpy ndarray.

    data:
        _nbytes (int)    : the bytes of AscendArray's data
        _shape (tuple)   : the shape of this array
        _dtype (int)     : the acl data type of AscendArray
        _flag (int)      : the flag of defined memory malloc on device/host/dvpp
        _mem (instance)  : save the instance of class Memory()
        _data (a pointer): the mem ptr of AscendArray's data

    Attributes:
        ascend_data (read, write) : get(read) the pointer of data malloced by itself(_data's value), 
                                    or band it to a new memory.
        shape (read, write)       : Tuple of array dimensions. (should work with write ascend_data).
        ndim (read, write)        : Number of array dimensions.
        strides	                  : Tuple of bytes to step in each dimension when traversing an array.
        dtype (read, write)       : get or write the data type.
        size (read)               : Number of elements in the array.
        nbytes (read, write)      : Total bytes consumed by the elements of the array.
        itemsize (read)           : Length of one AscendArray element in bytes, which equal to size * dtype(size).
        locate                    : 
        flag (read)               : Information about the memory layout of the array.

    Methods:
        reshape                   :
        resize                    :
        to_np                     : use this function to copy device data to host and formal like numpy array.
        to_ascend                 : trans a ndarray data to AscendArray(i.e. feed this instance with a numpy array data).
        clone                     : new a AscendArray object clone from np.ndarray
        to                        : copy this instance's data to another same shape AscendArray.
    """
    def __init__(self, shape, dtype, size=0, buffer=None, format=None, context=None, flag='DEVICE'):
        assert isinstance(shape, tuple), \
            f'Input shape of AscendArray instance expects tuple, but got {type(shape)}'

        try:
            self._dtype = np.dtype(dtype)
        except:
            raise TypeError(f'Input dtype expect a numpy.dtype, but got {type(dtype)}')
 
        self._shape = shape
        self._flag = flag
        self._context = context
        self._format = format

        bind_context(context)

        is_malloc = True
        # calc memory size according to variable calling
        # 1. initial a scalar
        #    Examples:
        #    --------
        #    >>> AscendArray((), dtype=np.int32)
        if shape == ():
            self._nbytes = self._dtype.itemsize

        # 2. initial a array with shape(2, 3): 
        #    Examples:
        #    --------
        #    >>> AscendArray((2, 3), dtype=np.float16)
        elif size <= 0 and buffer is None:
            self._nbytes = int(np.prod(shape) * self._dtype.itemsize)
        
        # 3. initial a array with shape(6,) and 256 bytes
        #    Examples:
        #    --------
        #    >>> AscendArray((6,), dtype=np.float32, size=256)
        elif buffer is None:
            self._nbytes = size
            self._shape = (shape[0], size//(shape[0] * self._dtype.itemsize))
 
        # 4. initial a array with shape(16,) and binding memory pointer mem_ptr
        #    Examples:
        #    --------
        #    >>> AscendArray((16,), dtype=np.float16, size=256, buffer=mem_ptr)
        else:
            self._nbytes = size
            self._shape = (shape[0], size//(shape[0] * self._dtype.itemsize))
            is_malloc = False

        # bind memory
        if is_malloc:
            self._mem = Memory(self._context, self._nbytes, flag)
            self._data = self._mem.ptr
        else:
            self._data = buffer
        
    # 1.ascend_data getter and setter
    @property
    def ascend_data(self):
        return self._data

    @ascend_data.setter
    def ascend_data(self, dev_ptr:int):
        assert isinstance(dev_ptr, int), \
            f'Function dev_ptr args of input expects int type, but got {type(dev_ptr)}'

        if hasattr(self, '_mem'):
            del self._mem
        self._data = dev_ptr

    # 2.memory location getter
    @property
    def flag(self):
        return self._flag

    # 3.context resource getter and setter
    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, context):
        """binding AscendArray with a new context.
            @note
            only support the memory of AscendArray create by itself.
        Args:
            context : the context to be binded.

        Returns:
            None.
        """
        assert isinstance(context, int), \
            f"Input context expects int value, but got {type(context)}."

        if not hasattr(self, '_mem'):
            raise ValueError(f"This AscendArray instance not support to set context.")
        else:
            del self._mem

        bind_context(context)
        self._mem = Memory(context, self._nbytes, self._flag)
        self._data = self._mem.ptr
    
    # 4.ndim getter
    @property
    def ndim(self):
        return len(self._shape)

    # 5.shape getter and setter
    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, shape):
        assert isinstance(shape, tuple), \
            f'Input shape expects a tuple, but got {type(shape)}'

        if len(shape) <= 0 or shape[0] == 0:
            raise ValueError('Input shape is empty or format is invalid in calling function shape')

        self._shape = shape

    # 6.dtype getter and setter
    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        try: 
            self._dtype = np.dtype(dtype)
        except:
            raise TypeError(f'Input dtype expects a numpy.dtype, but got {type(dtype)}')


    # 7.nbytes getter
    @property
    def nbytes(self):
        return self._nbytes

    # 8.itemsize getter
    @property
    def itemsize(self):
        return self._dtype.itemsize

    # 9.data format
    @property
    def format(self):
        return self._format

    @format.setter
    def format(self, format):
        if not isinstance(format, (str, int)):
            raise TypeError(f"Input format expects int or string, but got {type(format)}.")

        self._format = format

    @property
    def size(self):
        return np.prod(self._shape)

    def astype(self, dtype):
        """ Cast a tensor form src data type to dst data type. Firstly, we try to use Cast operator 
            to do. If it fails, we use numpy astype method. 
        Args:
            dtype : the shape to be reshaped.

        Returns:
            AscendArray : dst data.
        """
        try:
            from ..ops.op import Cast
            return Cast(self, dtype=dtype, context=self.context).data
        except:
            array = self.to_np.astype(dtype)
            return self.clone(array)

    def reshape(self, shape):
        """only reshape the view of AscendArray, and the data is original.
            @note
            only modify the view of AscendArray.
        Args:
            shape : the shape to be reshaped.

        Returns:
            AscendArray : the original object.
        """
        assert isinstance(shape, tuple), \
            f'Input shape expects tuple type, but got {type(shape)}.'
        
        assert np.prod(self._shape) == np.prod(shape), \
            f"The given shape({shape})'s elements should same to {self._shape}."
        
        self._shape = shape
        return self
    
    def resize(self, shape:tuple):
        """resize the shape and data of AscendArray.
            @note
            the data arrangement of AscendArray is modified.
        Args:
            shape : the resized new shape.

        Returns:
            AscendArray : the original object.
        """
        assert isinstance(shape, tuple), \
            f'Input args of func reshape expects tuple type, but got {type(shape)}.'

        pass


    def to_numpy(self, nbytes=None):
        """Copy the attributes and data of AscendArray to np.ndarry object.
        Args:
            None.

        Returns:
            np.ndarray object

        Usage:
            out = out_data.to_numpy()
        """
        if self._data is None:
            raise ValueError('Variable self._data is None in calling function to_np, \
                maybe this AscendArray instance parameter is null.')

        if nbytes and nbytes > self._nbytes:
            raise ValueError(f"Input nbytes must lower than {self._nbytes}, but got {nbytes}.")

        if self._flag == 'HOST':
            raise ValueError(f'Method to_np must be used with DEVICE or DVPP memory')
        
        _nbytes = nbytes if nbytes else self._nbytes
        array = np.zeros(_nbytes, dtype=np.byte)
        array_ptr = acl.util.numpy_to_ptr(array)

        # copy device data to numpy array_ptr
        memcpy_d2h(array_ptr, self._data, _nbytes)

        if self._dtype == np.dtype('float32'):
            form = str(_nbytes // 4) + 'f'
        elif self._dtype == np.dtype('uint8'):
            form = str(_nbytes) + 'B'
        elif self._dtype == np.dtype('int8'):
            form = str(_nbytes) + 'b'
        elif self._dtype == np.dtype('int32'):
            form = str(_nbytes // 4) + 'i'
        else:
            raise TypeError("Convert AscendArray data to numpy only support [float32, uint8, int8, int32].")

        tuple_st = struct.unpack(form, bytearray(array))
        array = np.array(tuple_st).astype(self._dtype).reshape(self._shape)

        return array
    
    @property
    def to_np(self):
        return self.to_numpy()

    def to_ascend(self, array):
        """ Copy all the data of array(np.ndarray) to AscendArray.
        Args:
            array : input np.ndarray

        Returns:
            None.

        Use like:
            array = np.ones(shape=(384, 384), dtype='float16')
            ascend_array = AscendArray(shape=(384, 384), nbytes=array.nbytes, dtype=NPY_USHORT)

            @notes: copy data of array to ascend_array
            ascend_array.to_ascend(array)
        """
        if self._data is None:
            raise ValueError('instance arg self._data is None in calling function to_ascend, \
                Maybe this AscendArray instance parameter is null.')

        if self._flag == 'HOST':
            raise ValueError(f'Method to_ascend only be used with DEVICE or DVPP memory')

        assert isinstance(array, np.ndarray), \
            f'Function to_ascend args of input expects a np.ndarray object, but got {type(array)}'

        assert (array.shape == self._shape) and (array.nbytes == self._nbytes), \
            'Function to_ascend of input expects same shape and nbytes,' \
            f' but actually we got shape:{array.shape}, nbytes:{array.nbytes}.'

        # get the array pointer for copy data to device 
        array_ptr = acl.util.numpy_to_ptr(array)

        # do copy
        memcpy_h2d(self._data, array_ptr, self._nbytes)

    @classmethod
    def clone(cls, array, context=None, format=None, flag="DEVICE"):
        """New a AscendArray instance and Clone all the attributes of np.ndarray to it.
        Args:
            array : A np.ndarray data to be cloned

        Returns:
            An AscendArray object.
        """
        assert isinstance(array, np.ndarray), \
            f'Input args array expects class np.ndarray object, but got {type(array)}.'

        if context and not isinstance(context, int):
            raise TypeError(f"Input context expects int type, but got {type(context)}.")
        
        bind_context(context)

        # get the array pointer for copy device data to host
        array_ptr = acl.util.numpy_to_ptr(array)

        # new an AscendArray object shape like input array.
        cloned_array = cls(shape=array.shape, dtype=array.dtype, format=format, context=context, flag=flag)

        # do copy
        memcpy_h2d(cloned_array.ascend_data, array_ptr, array.nbytes)

        return cloned_array

    def to(self, ascendarray):
        """ Copy a AscendArray data to another ascendarray(AscendArray).
        Args:
            ascendarray : A AscendArray to copy

        Returns:
            None.
        """
        if self._data is None:
            raise ValueError("Variable self._data is None in calling function 'to', \
                Maybe this AscendArray instance parameter is null.")

        assert isinstance(ascendarray, AscendArray), \
                f"Input args of func 'to' expects a class of AscendArray, but got {type(array)}."
 
        assert (self._nbytes <= ascendarray.nbytes) and (self._dtype == ascendarray.dtype), \
                "Shape or dtype of the input AscendArray is different from original."
        
        memcpy_d2d(ascendarray.ascend_data, self._data, self._nbytes)
 
    
    def __len__(self):
        """ Number of elements in the array, same to self.size.
        Args:
            None

        Returns:
            number of elements
        """
        return np.prod(self._shape)

    def __getitem__(self, idx):
        """ get AscendArray data using subscript index
        Args:
            idx : an int or slice object

        Returns:
            data of AscendArray
        """
        if self.dtype not in [
            np.dtype('float32'),
            np.dtype('int8')   ,
            np.dtype('int32')  ,
            np.dtype('uint8')  
        ]:
            raise TypeError("Only dtype in ['float32', 'int32', 'int8', 'uint8'] are support \
                            to using subscript index.")

        if not hasattr(self, '_cloned_array'):
            self._cloned_array = self.to_np
            self._cloned_array.reshape(self._shape)

        if isinstance(idx, int): 
            if idx < self.size and idx >= 0:
                return self._cloned_array[idx]
            elif idx > -self.size and idx < 0:
                return self._cloned_array[idx + self.size]
            else:
                raise IndexError(f"index {idx} is out of bounds for axis 0")

        elif isinstance(idx, slice):
            return self._cloned_array[idx]
        else:
            return 'index error'

    def __setitem__(self, index, value):
        """ release to set AscendArray data using subscript index
        Args:
            idx : an int or slice object

        Returns:
            data of AscendArray
        """
        if not hasattr(self, '_cloned_array'):
            self._cloned_array = self.to_np

        if isinstance(index, int) and index < self.size:
            self._cloned_array[index] = value
        elif isinstance(index, (list, tuple)):
            for i in index:
                assert i < self.size, f'index out of range.'
            if isinstance(value, (list, tuple)):
                if len(index) == len(value):
                    for i, v in enumerate(index):
                        self._cloned_array[v] = value[i]
                else:
                    raise Exception('values and index must be of the same length')
            elif isinstance(value, (int, float, str)):
                for i in index:
                    self._cloned_array[i] = value
            else:
                raise Exception('value error')
        else:
            raise Exception('index error')

        #copy to device
        self.to_ascend(self._cloned_array)

    def __del__(self):
        if hasattr(self, '_mem'):
            del self._mem
            self._data = None
        elif hasattr(self, '_data') and self._data is not None:
            free(self._data, flag=self._flag)

if __name__ == "__main__":
    import pdb
    from resource.context import Context
    resource = Context({12})
    size = 1382400
    ptr, ret = acl.media.dvpp_malloc(size)
    pdb.set_trace()
    array = AscendArray((1280,), np.dtype('uint8'), size=size, buffer=ptr, flag='DEVICE')

    array = np.ones(shape=(384, 384), dtype='float32')
    ascend_array = AscendArray.clone(array)
    print(ascend_array[2])
    print(ascend_array[:32])
    if isinstance(ascend_array, AscendArray):
        print("clone ndarray success.")
    out = ascend_array.to_np
    print(out)
    del resource